package org.jutils.log;
/*------------------------------------------------------------------------------
Name:      LogDeviceConsole.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   The console as logging device
------------------------------------------------------------------------------*/

import org.jutils.JUtilsException;


/******************************************************************************
This class implements the console as device for the logging architecture
from LogManager and LogChannel.
<p />
This log device does work synchronous.

@author  Juergen Freidling
******************************************************************************/
public class LogDeviceConsole
   implements LogConstants, LogableDevice
{
   // colors foreground/background (for xterm)
   private static final String ESC          = "\033[0m"; // Reset color to original values
   private static final String BOLD         = "\033[1m";

   private static final String RED_BLACK    = "\033[31;40m";
   private static final String GREEN_BLACK  = "\033[32;40m";
   private static final String YELLOW_BLACK = "\033[33;40m";
   private static final String BLUE_BLACK   = "\033[34;40m";
   private static final String PINK_BLACK   = "\033[35;40m";
   private static final String LTGREEN_BLACK= "\033[36;40m";
   private static final String WHITE_BLACK  = "\033[37;40m";

   private static final String WHITE_RED    = "\033[37;41m";
   private static final String BLACK_RED    = "\033[30;41m";
   private static final String BLACK_GREEN  = "\033[40;42m";
   private static final String BLACK_PINK   = "\033[40;45m";
   private static final String BLACK_LTGREEN= "\033[40;46m";

   // Output text for different logging levels
   private static final String errorX = new String("ERROR");
   private static final String warnX  = new String("WARN ");
   private static final String infoX  = new String("INFO ");
   private static final String callX  = new String("CALL ");
   private static final String timeX  = new String("TIME ");
   private static final String traceX = new String("TRACE");
   private static final String dumpX  = new String("DUMP ");
   private static final String plainX = new String("     ");

   private static final String errorE = new String(RED_BLACK    + "ERROR" + ESC);
   private static final String warnE  = new String(YELLOW_BLACK + "WARN " + ESC);
   private static final String infoE  = new String(GREEN_BLACK  + "INFO " + ESC);
   private static final String callE  = new String(BLUE_BLACK   + "CALL " + ESC);
   private static final String timeE  = new String(LTGREEN_BLACK+ "TIME " + ESC);
   private static final String traceE = new String(WHITE_BLACK  + "TRACE" + ESC);
   private static final String dumpE  = new String(PINK_BLACK   + "DUMP " + ESC);
   private static final String plainE = new String(WHITE_BLACK  + "     " + ESC);


   private LogChannel channel = null;
   // Colored output to xterm
   private boolean withXtermEscapeColor = false;




   /***************************************************************************
    * Constructor of this object. It automatically call the basic
    * initialization.
    *
    * @param      channel
    *             the LogChannel object which is the owner of this object.
    **************************************************************************/
   public LogDeviceConsole(LogChannel channel) {
      this.channel = channel;
      initialize();
   }


   /***************************************************************************
    * Basic initialization of this object.
    **************************************************************************/
   public void initialize() {
      String osName = System.getProperty("os.name"); // "Linux" "Windows NT"...
      if (osName.startsWith("Window"))
         withXtermEscapeColor = false;
      else
         withXtermEscapeColor = true;
   }


   /***************************************************************************
    * Method to manually set the color mode of the console output.
    *
    * @param      bValue
    *             true if color mode should be activated, false if color mode
    *             should be deactivated.
    **************************************************************************/
   public void setColorMode(boolean bValue) {
      withXtermEscapeColor = bValue;
   }


   /***************************************************************************
    * Method to get the actual setting of the color mode.
    *
    * @return     true if color mode is set, false otherwise.
    **************************************************************************/
   public boolean getColorMode() {
      return withXtermEscapeColor;
   }


   /***************************************************************************
    * Internal method for converting int level values to string
    * representation. This method also sets the xterminal color codes to the
    * string, if color mode is activated.
    *
    * @param      level
    *             the int value to convert
    * @return     the resulting string, maybe with color escape sequences,
    *             or the string "UNKNOWN LEVEL x", if given level is unknown.
    **************************************************************************/
   private String convertLevelToString(int level) {
      if ((level & LOG_ERROR) > 0) {
         if (withXtermEscapeColor == true)
            return errorE;
         else
            return errorX;
      }
      if ((level & LOG_WARN) > 0) {
         if (withXtermEscapeColor == true)
            return warnE;
         else
            return warnX;
      }
      if ((level & LOG_INFO) > 0) {
         if (withXtermEscapeColor == true)
            return infoE;
         else
            return infoX;
      }
      if ((level & LOG_CALL) > 0) {
         if (withXtermEscapeColor == true)
            return callE;
         else
            return callX;
      }
      if ((level & LOG_TIME) > 0) {
         if (withXtermEscapeColor == true)
            return timeE;
         else
            return timeX;
      }
      if ((level & LOG_TRACE) > 0) {
         if (withXtermEscapeColor == true)
            return traceE;
         else
            return traceX;
      }
      if ((level & LOG_DUMP) > 0) {
         if (withXtermEscapeColor == true)
            return dumpE;
         else
            return dumpX;
      }
      return "UNKNOWN LEVEL" + level;
   }


   /***************************************************************************
    * The logging method, which displays the logging data at the console
    *
    * @param      level
    *             the debug level, which the data represents, e.g.
    *             LogChannel.LOG_ERROR
    * @param      source
    *             a unique identifier, to determine who is logging this data.
    * @param      text
    *             the real data to log
    **************************************************************************/
   public void log(int level, String source, String text) {
      if (channel == null) {
         return;
      }
      String levelString = convertLevelToString(level);
      String logEntry = channel.formatLogData(levelString, level, source, text);
      if ((level & LOG_ERROR) != 0 || (level & LOG_WARN) != 0) {
         System.err.println(logEntry);
      }
      else {
         System.out.println(logEntry);
      }
   }
}
