package org.jutils.log;
/*------------------------------------------------------------------------------
Name:      LogDeviceFile.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   A file as logging device
------------------------------------------------------------------------------*/

import java.io.File;
import java.io.FileOutputStream;

import org.jutils.JUtilsException;


/******************************************************************************
This class implements the file as device for the logging architecture
from LogManager and LogChannel.
<p />
This log device does work synchronous.

@author  Juergen Freidling
******************************************************************************/
public class LogDeviceFile
   implements LogConstants, LogableDevice
{
   private static String ME = "LogDeviceFile";

   private static final String errorX = new String("ERROR");
   private static final String warnX  = new String("WARN ");
   private static final String infoX  = new String("INFO ");
   private static final String callX  = new String("CALL ");
   private static final String timeX  = new String("TIME ");
   private static final String traceX = new String("TRACE");
   private static final String dumpX  = new String("DUMP ");
   private static final String plainX = new String("     ");


   private LogChannel channel = null;
   private String fileName = "jutils.log";
   private File logFile = null;
   private FileOutputStream to = null;
   private int maxLogFileLines = 50000; // max # of lines per file
   private int numLogLines = 0;
   private byte[] newLine = null;
   private int fileCounter = 1;




   /***************************************************************************
    * Constructor of this object. It automatically call the basic
    * initialization.
    *
    * @param      channel
    *             the LogChannel object which is the owner of this object.
    * @param      fileName
    *             the name of the destination file, to which the logging
    *             data should be logged.
    **************************************************************************/
   public LogDeviceFile(LogChannel channel, String fileName) {
      this.channel = channel;
      if (fileName != null && fileName.length() > 1)
         this.fileName = fileName;
      initialize();
   }


   /***************************************************************************
    * Basic initialization of this object.
    **************************************************************************/
   public void initialize() {
      // set the line seperator to the OS specific value
      String ls = System.getProperty("line.separator");
      newLine = ls.getBytes();
      try {
         this.to = new FileOutputStream(fileName, true);
      }
      catch (Exception e) {
         // on error this device is removed from LogChannel object, so
         // that no further errors are reported!
         channel.removeLogDevice(this);
         channel.error(ME, "Can't access file " + fileName + ": " + e.toString());
         channel.info(ME, "Removing logging device for " + fileName);
      }
   }


   /***************************************************************************
    * Convert the given level value to a readable string
    *
    * @param      level
    *             the given debug level which to convert.
    * @return     the debug level as readable String
    **************************************************************************/
   private String convertLevelToString(int level) {
      if ((level & LOG_ERROR) > 0) {
         return errorX;
      }
      if ((level & LOG_WARN) > 0) {
         return warnX;
      }
      if ((level & LOG_INFO) > 0) {
         return infoX;
      }
      if ((level & LOG_CALL) > 0) {
         return callX;
      }
      if ((level & LOG_TIME) > 0) {
         return timeX;
      }
      if ((level & LOG_TRACE) > 0) {
         return traceX;
      }
      if ((level & LOG_DUMP) > 0) {
         return dumpX;
      }
      return "UNKNOWN LEVEL" + level;
   }


   /***************************************************************************
    * Internal method to swap the logfiles if the maximum number of lines
    * per logfile is reached.
    **************************************************************************/
   private void newLogFile() {
      // first close existing output stream
      try {
         to.close();
      }
      catch (Throwable t) {
      }

      // backup file
      String bakFile = fileName + ".backup" + fileCounter;
      fileCounter++;
      File nf = new File(bakFile);
      if (nf.exists())
         nf.delete();
      logFile = new File(fileName);
      if (logFile.exists()) {
         logFile.renameTo(nf);
      }

      // new logfile
      logFile = new File(fileName);
      if (logFile.exists()) {
         logFile.delete();
      }
      numLogLines = 0;

      // set up new output stream
      try {
         this.to = new FileOutputStream(fileName, true);
      }
      catch (Exception e) {
         // on error this device is removed from LogChannel object, so
         // that no further errors are reported!
         channel.removeLogDevice(this);
         channel.error(ME, "Can't access file " + fileName + ": " + e.toString());
         channel.info(ME, "Removing logging device for " + fileName);
      }
   }


   /***************************************************************************
    * The logging method, which sends the logging data to the file.
    *
    * @param      level
    *             the debug level, which the data represents, e.g.
    *             LogChannel.LOG_ERROR
    * @param      source
    *             a unique identifier, to determine who is logging this data.
    * @param      text
    *             the real data to log
    **************************************************************************/
   public void log(int level, String source, String text) {
      if (channel == null) {
         return;
      }
      String levelString = convertLevelToString(level);
      String logEntry = channel.formatLogData(levelString, level, source, text);
      if (numLogLines > maxLogFileLines)
         newLogFile();
      numLogLines++;
      try {
         to.write(logEntry.getBytes());
         to.write(newLine);
      }
      catch (Exception e) {
         // on error this device is removed from LogChannel object, so
         // that no further errors are reported!
         channel.removeLogDevice(this);
         channel.error(ME, "Can't write to file " + fileName + ": " + e.toString());
         channel.info(ME, "Removing logging device for " + fileName);
      }
   }
}
